/*
 * mboxa.c
 *
 * Created: 2013/11/04 14:12:49
 *  Author: Chuck
 */ 


#include <avr/io.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include "config/midi_conf.h"
#include "global.h"
#include "uart.h"
#include "midi.h"

#define CC_CYCLE 46
#define CC_PLAY 42
#define CC_STOP 41

#define TAP_TIMER_START (TCNT0 = 0, beat_cnt = 0, TCCR0B |= (1<<CS01))
#define TAP_TIMER_STOP  (TCCR0B &= ~(1<<CS01))
#define F8_TIMER_START  (TCNT1 = 0, oddeven = 0, TCCR1B |= (1<<CS11))
#define F8_TIMER_STOP   (TCCR1B &= ~(1<<CS11))
#define MIDI_START      (PORTB |= (1<<PB4))
#define MIDI_STOP       (PORTB &= ~(1<<PB4))
#define F8_RAISE        (PORTD |= (1<<PD6))
#define F8_FALL         (PORTD &= ~(1<<PD6))
#define F8_CNT OCR1A

enum { IDLE, THREE, FOUR, STABLE };

struct MIDI_SystemMessage MidiSystemMessage;
struct MIDI_ChannelMessage MidiChannelMessage;

static uint8_t midich_rx;
static uint8_t TimedOut = DISABLE;
static uint16_t beat_cnt = 0;
static uint8_t oddeven = 0;


ISR(TIMER0_COMPA_vect)
{
	beat_cnt++;
}

ISR(TIMER1_COMPA_vect)
{
	// process once every 2 interrupts
	if (oddeven & 0x01) {
		F8_RAISE;
	}
	oddeven++;
	F8_FALL;
}


void timer0Init(void)
{
	// start timer0
	//TCCR0A
	//TCCR0B |= (1<<CS01) ck/8 cnt *64 -> 512ck = 25.6usec tick 
	//TCCR0B |= (1<<CS01) ck/8 cnt *48 -> 384ck = 19.2usec tick

	TCNT0 = 0L;
	OCR0A = 48;
	TCCR0A |= (1<<WGM01); 
	TIMSK |= (1<<OCIE0A);
}


void timer1Init(void)
{
	// start timer1
	//TCCR1A
	//TCCR1B  drive ck/16 (cant select) -> ck/8 (alternative)

	TCNT0 = 0L;
	OCR1A = 26042;
	TCCR1B |= (1<<WGM12);
	TIMSK |= (1<<OCIE1A);
}


static void ioinit (void)
{
	// input settings
	PORTB = 0x0F;
	DDRB |= (1<<PB4);	// port B: input w/ pull-up
	PORTD = 0xFE; // PD0 as input w/ pull-up
	DDRD |= (1<<PD6); // 

	// no timer user
	//	TCCR0B = (1<<CS00);
	//	TCCR1B = (1<<WGM12)|(1<<CS10);
	//	TIMSK = (1<<TOIE1);

	//	wdt_enable(WDTO_4S);
}

static void MyMidiInit(void)
{
	uint8_t tmpb;

	tmpb = ~PINB;
	midich_rx = tmpb & 0x0F;

	MidiSystemMessage.Status = 0;
	MidiChannelMessage.Status = 0;
}


int main(void)
{
	static uint8_t stat = IDLE;
	static uint8_t IsTapped = DISABLE;

	ioinit();
	MidiInit();
	MyMidiInit();
	timer0Init();
	timer1Init();
	
	sei();
    while(1)
    {
		MidiParse();
		if (MidiChannelMessage.Status & fMidiControlChange) {
			if (MidiChannelMessage.Channel == midich_rx) {
				if (MidiChannelMessage.ControlChange.Control==CC_CYCLE) {
					if (MidiChannelMessage.ControlChange.Value)
						IsTapped = ENABLE;
				} else if (MidiChannelMessage.ControlChange.Control==CC_PLAY) {
					if (MidiChannelMessage.ControlChange.Value) {
						//MidiSystemMessage.Status |= fMidiStart;
						//MidiSendSystemMsg();
						MIDI_START;
					}					
				} else if (MidiChannelMessage.ControlChange.Control==CC_STOP) {
					if (MidiChannelMessage.ControlChange.Value) {
						//MidiSystemMessage.Status |= fMidiStop;
						//MidiSendSystemMsg();
						MIDI_STOP;
					}					
				}
			}
			MidiChannelMessage.Status &= ~fMidiControlChange;
		}

		switch (stat) {
		case IDLE:
			// counter clear
			if (IsTapped) {
				stat = THREE;
				TAP_TIMER_START;
				IsTapped = DISABLE;
			}
			break;
		case THREE:
			if (IsTapped) {
				stat = FOUR;
				TAP_TIMER_STOP;
				F8_RAISE;
				F8_CNT = beat_cnt;
				F8_TIMER_START;
				IsTapped = DISABLE;
				F8_FALL;
			} else if (beat_cnt==65535) { // timer overflow
				stat = IDLE;
				TAP_TIMER_STOP;
				beat_cnt = 0;
			}
			break;
		case FOUR:
			stat = STABLE;
			break;
		case STABLE:
			// counter clear
			if (IsTapped) {
				F8_TIMER_STOP;
				stat = THREE;
				IsTapped = DISABLE;
				TAP_TIMER_START;
			}
			break;
		default:
			stat = IDLE;
		}


    }
}