/*!
	@brief	MIDI parsing library

	@file	midi.c
	@author	Satoshi "Chuck" Takagi
	@date	2009/04/03
 */


#include <inttypes.h>
#include "midi.h"

static uint8_t _midiSystemMessage(void);
static void _midiChannelMessage(void);

static uint8_t rb;				//!< read byte from UART
#ifdef NOUSE
static uint8_t midich_rx = 0;	//!< MIDI channel
static uint8_t midich_tx = 0;	//!< MIDI channel
#endif // NOUSE
static uint8_t midi_stat = 255;	//!< MIDI parsing status initial value
static uint8_t midi_ch;			//!< MIDI receive channel from stream
static uint8_t midi_byte[3];	//!< MIDI parsing work array


//	@brief	MIDI VA|[g̏s
void MidiInit(void)
{
	midiUartInit();
}

#ifdef NOUSE
//	@brief	MIDI channel ԍL^
void MidiSetChannel_RX(uint8_t ch)
{
	if (ch < 16)
		midich_rx = ch;
}


//	@brief	MIDI channel ԍL^
void MidiSetChannel_TX(uint8_t ch)
{
	if (ch < 16)
	midich_tx = ch;
}


//	@brief	MIDI channel ԍo
uint8_t MidiGetChannel_RX(void)
{
	return (midich_rx);
}

//	@brief	MIDI channel ԍo
uint8_t MidiGetChannel_TX(void)
{
	return (midich_tx);
}
#endif // NOUSE

//	@brief	MIDI Parser
/*!
	@brief	MIDI Parser

	@param	Ȃ

	@note	MIDIVXebZ[W͂Asystem exclusive͏OĂȊO̕KvȃbZ[W\̂ɋL^Bmain loop ŒIɌĂяoƁB
*/

void MidiParse(void)
{
	int16_t c;

    rb = c = midiUartGetByte();
    if(c != -1){
		if( _midiSystemMessage() ){
			_midiChannelMessage();
		}
	}
}



/*!
	@brief	MIDI System Message parser

	@param	Ȃ
	@return	MIDI Channel Message1

	@note	MIDIVXebZ[W͂Asystem exclusive͏OĂȊO̕KvȃbZ[WMidiSystemMessage\̂ɋL^B
*/
static uint8_t _midiSystemMessage(void)
{
	static uint8_t sysex;

	if (sysex) {					// in sysex process
		if (rb == 0xF7) sysex=0;	// end of sysex process
	} else {
		if (rb < 0xF8) { 
			if (rb >= 0xF0) {		// common msg
				if (rb == 0xF0)	sysex = 1;	// start of sysex process
				midi_stat = 100;
			} else if (rb > 0x7F) {	// voice msg
				midi_ch = rb & 0x0F;
				//if (midi_ch == midich_rx) { // for multi-channel recieve
					midi_stat = 0;
				//}
			} else {				// store midi data byte 00-7F
				midi_byte[midi_stat & 0x01] = rb;
 			}
 			return 1;
 		} else {					// realtime msg >= 0xF8
			switch (rb) {
			case 0xF8:	MidiSystemMessage.Status |= fMidiTimingClock;	break;		// timing clock
			case 0xF9:	break;
			case 0xFA:	MidiSystemMessage.Status |= fMidiStart;			break;		// start
			case 0xFB:	MidiSystemMessage.Status |= fMidiContinue;		break;		// continue
			case 0xFC:	MidiSystemMessage.Status |= fMidiStop;			break;		// stop
			case 0xFD:	break;
			case 0xFE:	MidiSystemMessage.Status |= fMidiActiveSensing; break;		// active sensing
			case 0xFF:	MidiSystemMessage.Status |= fMidiSystemReset;   break;		// system reset
			}
		}
	}

	return 0;
}

/*!
	@brief	MIDI Channel Message parser

	@param	Ȃ
	@return	Ȃ

	@note	MIDI`lbZ[W͂AKvȃbZ[WMidiChannelMessage\̂ɋL^B1̃R[ɂ1oCgAiԓmidi_byteɂĊǗĂB
*/
static void _midiChannelMessage(void)
{
	switch (midi_stat) {
	case 0:
		switch ( rb & 0xF0 ) {
#ifdef USE_MIDI_NOTE
		case MIDI_NOTE_OFF			: midi_byte[2] = rb; midi_stat=10;  break;
		case MIDI_NOTE_ON			: midi_byte[2] = rb; midi_stat=20;  break;
#endif // USE_MIDI_NOTE

#ifdef USE_MIDI_PBEND
		case MIDI_PBEND				: midi_byte[2] = rb; midi_stat=30; break;
#endif // USE_MIDI_PBEND

#ifdef USE_MIDI_CC
		case MIDI_CC				: midi_byte[2] = rb; midi_stat=40; break;
#endif // USE_MIDI_CC

#ifdef USE_MIDI_PC
		case MIDI_PC				: midi_byte[2] = rb; midi_stat=50;  break;
#endif // USE_MIDI_PC

#ifdef NO_USE
		case MIDI_POLY_PRESSURE		: midi_byte[2] = rb; midi_stat=60;  break;
		case MIDI_CHANNEL_PRESSURE	: midi_byte[2] = rb; midi_stat=70; break;
#endif
		default:
			midi_stat=255;
		}
		break;

#ifdef USE_MIDI_NOTE
	/* Note Off */
	case 10: midi_stat=11; break;
	case 11: midi_stat=10;
			//if (midi_ch == midich_rx) {
				MidiChannelMessage.Channel = midi_ch;
				MidiChannelMessage.NoteOff.Note = midi_byte[0];
				MidiChannelMessage.NoteOff.Velocity = midi_byte[1];
				MidiChannelMessage.Status |= fMidiNoteOff;
			//}
			break;

	/* Note On */
	case 20: midi_stat=21; break;
	case 21: midi_stat=20;
			//if (midi_ch == midich_rx) {
				MidiChannelMessage.Channel = midi_ch;
				if (midi_byte[1] == 0) {
					// note off
					MidiChannelMessage.NoteOff.Note = midi_byte[0];
					MidiChannelMessage.NoteOff.Velocity = 64;
					MidiChannelMessage.Status |= fMidiNoteOff;
				} else {
					MidiChannelMessage.NoteOn.Note = midi_byte[0];
					MidiChannelMessage.NoteOn.Velocity = midi_byte[1];
					MidiChannelMessage.Status |= fMidiNoteOn;
				}
			//}
			break;
#endif // USE_MIDI_NOTE

#ifdef USE_MIDI_PBEND
	/* PitchBend */
	case 30: midi_stat=31; break;
	case 31: midi_stat=30;
			//if (midi_ch == midich_rx) {
				MidiChannelMessage.Channel = midi_ch;
				MidiChannelMessage.PitchBend.Value = (midi_byte[1]<<7) | midi_byte[0];
				MidiChannelMessage.Status |= fMidiPitchBend;
			//}
			break;
#endif // USE_MIDI_PBEND

#ifdef USE_MIDI_CC
	/* Control Change */
	case 40: midi_stat=41; break;
	case 41: midi_stat=40;			// data and for running status
			//if (midi_ch == midich_rx) {
				MidiChannelMessage.Channel = midi_ch;
				MidiChannelMessage.ControlChange.Control = midi_byte[0];
				MidiChannelMessage.ControlChange.Value = midi_byte[1];
				if (midi_byte[0] < 120)
					MidiChannelMessage.Status |= fMidiControlChange;
				else
					MidiChannelMessage.Status |= fMidiModeChange;
			//}
			break;
#endif // USE_MIDI_CC

#ifdef USE_MIDI_PC
	/* Program change */
	case 50:
			//if (midi_ch == midich_rx) {
				MidiChannelMessage.Channel = midi_ch;
				MidiChannelMessage.ProgramChange.Value = midi_byte[0];
				MidiChannelMessage.Status |= fMidiProgramChange;
			//}
			break;

#endif // USE_MIDI_PC

#ifdef NO_USE
	/* Polyphonic Key Pressure */ 
	case 60: midi_stat=61; break;
	case 61: midi_stat=60;
			//if (midi_ch == midich_rx) {
				MidiChannelMessage.Channel = midi_ch;
				MidiChannelMessage.PolyPressure.Note = midi_byte[0];
				MidiChannelMessage.PolyPressure.Value = midi_byte[1];
				MidiChannelMessage.Status |= fMidiPolyPressure;
			//}
			break;

	/* ChannelPressure */
	case 70:
			//if (midi_ch == midich_rx) {
				MidiChannelMessage.Channel = midi_ch;
				MidiChannelMessage.ChannelPressure.Value = midi_byte[0];
				MidiChannelMessage.Status |= fMidiChannelPressure;
			//}
			break;
#endif // NO_USE

#ifdef NOT_IMPLEMENTED
	case 100:
		switch ( rb ) {
		case MIDI_MIDITC   : midi_stat = 110; break;
		case MIDI_SONG_PTR : midi_stat = 120; break;
		case MIDI_SONG_SEL : midi_stat = 130; break;
		case MIDI_TUNE_REQ : midi_stat = 255; break;
		case 0xF4          : midi_stat = 255; break;
		case 0xF5          : midi_stat = 255; break;
		}
		break;

	/* MIDI timecode */
	case 110: break;

	/* Song Position Pointer */
	case 120: midi_stat=121; break;
	case 121: break;

	/* Song Select */
	case 130: break;
#endif // NOT_IMPLEMENTED
	
	default: break;
	}

}

void MidiSendSystemMsg(void)
{
	if (MidiSystemMessage.Status & fMidiTimingClock) {
		midiUartSendByte(0xF8);
		MidiSystemMessage.Status &= ~fMidiTimingClock;
	}	
	if (MidiSystemMessage.Status & fMidiStart) {
		midiUartSendByte(0xFA);
		MidiSystemMessage.Status &= ~fMidiStart;
	}
	if (MidiSystemMessage.Status & fMidiContinue) {
		midiUartSendByte(0xFB);
		MidiSystemMessage.Status &= ~fMidiContinue;
	}	
	if (MidiSystemMessage.Status & fMidiStop) {
		midiUartSendByte(0xFC);
		MidiSystemMessage.Status &= ~fMidiStop;
	}
	if (MidiSystemMessage.Status & fMidiActiveSensing) {
		midiUartSendByte(0xFE);
		MidiSystemMessage.Status &= ~fMidiActiveSensing;
	}	
	if (MidiSystemMessage.Status & fMidiSystemReset) {
		midiUartSendByte(0xFF);
		MidiSystemMessage.Status &= ~fMidiSystemReset;
	}	
}

void MidiSendChannelMsg(void)
{

	/* Note Off */	
	if ( MidiChannelMessage.Status & fMidiNoteOff ) {
		midiUartSendByte( MIDI_NOTE_OFF | MidiChannelMessage.Channel );
		midiUartSendByte( MidiChannelMessage.NoteOff.Note );
		midiUartSendByte( MidiChannelMessage.NoteOff.Velocity );
		MidiChannelMessage.Status &= ~fMidiNoteOff;
	}
	
	/* Note On */
	if ( MidiChannelMessage.Status & fMidiNoteOn ) {
		midiUartSendByte( MIDI_NOTE_ON | MidiChannelMessage.Channel );
		midiUartSendByte( MidiChannelMessage.NoteOn.Note );
		midiUartSendByte( MidiChannelMessage.NoteOn.Velocity );
		MidiChannelMessage.Status &= ~fMidiNoteOn;
	}

	/* PitchBend */
	if ( MidiChannelMessage.Status & fMidiPitchBend ) {
		uart_putchar( MIDI_PBEND | MidiChannelMessage.Channel );
		uart_putchar( (uint8_t)(MidiChannelMessage.PitchBend.Value & 0x007F) );
		uart_putchar( (uint8_t)((MidiChannelMessage.PitchBend.Value >>7) & 0x007F) );
		MidiChannelMessage.Status &= ~fMidiPitchBend;
	}

	/* Control Change */
	if ( MidiChannelMessage.Status & fMidiControlChange) {
		midiUartSendByte( MIDI_CC | MidiChannelMessage.Channel );
		midiUartSendByte( MidiChannelMessage.ControlChange.Control );
		midiUartSendByte( MidiChannelMessage.ControlChange.Value );
		MidiChannelMessage.Status &= ~fMidiControlChange;
	}

	/* Mode Change */
	if ( MidiChannelMessage.Status & fMidiModeChange) {
		midiUartSendByte( MIDI_CC | MidiChannelMessage.Channel );
		midiUartSendByte( MidiChannelMessage.ControlChange.Control );
		midiUartSendByte( MidiChannelMessage.ControlChange.Value );
		MidiChannelMessage.Status &= ~fMidiModeChange;
	}

	/* Mode Change */
	if ( MidiChannelMessage.Status & fMidiProgramChange) {
		midiUartSendByte( MIDI_PC | MidiChannelMessage.Channel );
		midiUartSendByte( MidiChannelMessage.ProgramChange.Value );
		MidiChannelMessage.Status &= ~fMidiProgramChange;
	}

	/* Polyphonic Key Pressure */
	if ( MidiChannelMessage.Status & fMidiPolyPressure) {
		midiUartSendByte( MIDI_POLY_PRESSURE | MidiChannelMessage.Channel );
		midiUartSendByte( MidiChannelMessage.PolyPressure.Note );
		midiUartSendByte( MidiChannelMessage.PolyPressure.Value );
		MidiChannelMessage.Status &= ~fMidiPolyPressure;
	}

	/* ChannelPressure */
	if ( MidiChannelMessage.Status & fMidiChannelPressure) {
		midiUartSendByte( MIDI_CHANNEL_PRESSURE | MidiChannelMessage.Channel );
		midiUartSendByte( MidiChannelMessage.ChannelPressure.Value );
		MidiChannelMessage.Status &= ~fMidiChannelPressure;
	}


	#ifdef NOT_IMPLEMENTED
		case MIDI_MIDITC   : midi_stat = 110; break;
		case MIDI_SONG_PTR : midi_stat = 120; break;
		case MIDI_SONG_SEL : midi_stat = 130; break;
		case MIDI_TUNE_REQ : midi_stat = 255; break;
		case 0xF4          : midi_stat = 255; break;
		case 0xF5          : midi_stat = 255; break;
	#endif	
}